#
#    File name: Fit_models.r
#
#    We fit four models
#
#    Model 1: Simple 2d smooth with no shocks
#    Model 2: 2d smooth with simple shocks (alpha = 0)
#    Model 3: 2d smooth with simple scaled shocks (alpha = 1)
#    Model 4: 2d smooth with optimal scaled shocks (alpha = 3.5)
#
#    ======================================================
#
#     Model 1: Smooth 2-d model - no shocks
#
#     This is done by using the Calc( ) function.  This does NOT fit
#     shocks and come from earlier work. It is used as a simple check on 
#     the general Calc.Shock( ) function which does allow shocks.
#
optBIC <- function(X){
  Calc(B.a, B.y, Rten.B.a, Rten.B.y, c.a, c.y, V, Off, Dth, A.init,
       P.a, X[1], P.y, X[2], 10^(-6), FALSE, 20, TRUE)$Bic
}
#
optlambda <- optim(c(1,1), optBIC, NULL, method = "L-BFGS-B", lower=c(0,0.001))
#
L.a.2d <- optlambda$par[1]
L.y.2d <- optlambda$par[2]
Ans.2d <- Calc(B.a, B.y, Rten.B.a, Rten.B.y, c.a, c.y, V, Off, Dth, A.init,
       P.a, L.a.2d, P.y, L.y.2d, 10^(-6), FALSE, 20, FALSE)
#
#   Here are the answers you should find (line 1 of Table 1)
#
#   Age.dx = 5, Year.dx = 5
#   c.a = 19, c.y = 23
#   Lambda.a = 10.24, Lambda.y = 6.50,
#   Tr = 284.84, Dev = 20918.24, Bic = 23484.56
#
#   ===========================================================
#
#     Model 2: 2-d model with shocks
#
#     Check that shock model with lambda.s = 10^12 is the same
#     as the 2d model with no shocks.
#
#     NB. We need one smoothing parameter per year - this is so we 
#     are ready for smoothing shocks taking account of shock size.
#
Lambda.s <- rep(10^12, nrow(B.y))
temp <- Calc.Shock(B.a, B.y, B.s, Rten.B.a, Rten.B.y, Rten.B.s, Rten.Ba.Bs,
                   c.a, c.y, c.s, V, Off, Dth, A.init, As.init,
                   P.a, L.a.2d, P.y, L.y.2d, Lambda.s,
                   10^(-6), TRUE, 10, TRUE)
#
#    and you can check this is the same as Ans.2d
#
#    I found optim( ) not very helpful so I used a simple grid search.
#    Here is the optimal answer I found
#
L.a.Shock <- 0.01
L.y.Shock <- 2150
L.s.Shock <- rep(800, nrow(B.y))
Ans.Shock <- Calc.Shock(B.a, B.y, B.s, Rten.B.a, Rten.B.y, Rten.B.s,
                        Rten.Ba.Bs, c.a, c.y, c.s, V, Off, Dth, A.init,
                        As.init, P.a, L.a.Shock, P.y, L.y.Shock, L.s.Shock,
                        10^(-8), FALSE, 10, TRUE)
#
#   Here are the answers you should find (line 4 of Table 1)
#
#   Age.dx = 5, Year.dx = 5, Shock.dx = 15
#   c.a = 19, c.y = 23, c.s = 9
#   Lambda.a = 0.01, Lambda.y = 2150, Lambda.s = 800
#   Tr = 454.7, Dev = 9354.4, Bic = 13450.90
#
#   ===================================================
#
#   Shock modelling: section 4.2 of the paper
#
#   Sensitive shock model with Lambda.s proportional to power
#   of 1/mean(abs(Shock))
#
#   Calculate smooth surface, fitted surface and shocks from Model 2
#
A.smooth <- Ans.Shock$A.smooth
A.shock <- Ans.Shock$A.shock
Ita <- Ans.Shock$Ita - Off
Ita.smooth <- B.a %*% matrix(A.smooth, nrow = ncol(B.a),
                              ncol = ncol(B.y)) %*% t(B.y)
Ita.shock <- Ita - Ita.smooth
#
Shock.Size <- function(x) mean(abs(x))
Size.of.Shocks <- apply(Ita.shock, 2, Shock.Size)
Order.of.Shocks <- order(Size.of.Shocks)
cbind(Year[Order.of.Shocks], Size.of.Shocks[Order.of.Shocks])
#
#   Set up scaling of smoothing parameter for shocks
#   This comes from equation(4.16)
#
Scale <- max(Size.of.Shocks)/Size.of.Shocks
#
#   =======================================================
#
#   Model 3: Simple scaling with alpha = 1
#   Again the optimal value is found with a grid search
#
Ans.Scale.1 <- Calc.Shock(B.a, B.y, B.s, Rten.B.a, Rten.B.y, Rten.B.s,
                        Rten.Ba.Bs, c.a, c.y, c.s, V, Off, Dth, A.init,
                        As.init, P.a, 0.01, P.y, 1300, 250*Scale,
                        10^(-8), FALSE, 10, TRUE)
#
#   Save fitted surface for smooth + simple shocks
#
Ita1 <- Ans.Scale.1$Ita - Off  
#
#   Here are the answers you should find (line 6 of Table 1)
#
#   Age.dx = 5, Year.dx = 5, Shock.dx = 15
#   c.a = 19, c.y = 23, c.s = 9
#   Lambda.a = 0.01, Lambda.y = 1300, Lambda.s = 250
#   Tr = 321.5, Dev = 9724.5, Bic = 12620.7
#
#   ==============================================================
#
#   Model 4:  Scaling optimizing over alpha
#
#   Again the optimal value is found with a grid search
#   (A bit laborious this - better numerical methods needed here!)
#
#   Here are some answers for given alpha with the optimal 
#   value of alpha being 3.5
#
#   Scale <- Scale^0
#   Answer: c(0.01, 2150, 800): Bic = 13451
#
#   Scale <- Scale^1
#   Answer: c(0.01, 1300, 250): Bic = 12621
#
#   Scale <- Scale^2
#   Answer: c(0.01, 1500, 25):  Bic = 12391
#
#   Scale <- Scale^3
#   Answer: c(0.01, 1600, 2.3): Bic = 12312
#
#   Scale <- Scale^3.5
#   Answer: c(0.01, 1600, 0.70): Bic = 12307 *****
#
#   Scale <- Scale^4
#   Answer: c(0.01, 1500, 0.2): Bic = 12318
#
Scale <- max(Size.of.Shocks)/Size.of.Shocks
Scale <- Scale^3.5
#
#    Figure 6: Plot of optimal shrinkage parameters (log_10)
#
par(mfrow = c(1,1))
plot(Year, log10(Scale), ylab = "Shrinkage parameters (log 10)",
     main = "Shrinkage parameters in scaled shock model")
abline(v = c(1918, 1919, 1923, 1944))
text(1969,0.5, "1918, 1919, 1923, 1944", adj = 0, cex = 1)
#
Ans.Scale <- Calc.Shock(B.a, B.y, B.s, Rten.B.a, Rten.B.y, Rten.B.s,
                        Rten.Ba.Bs, c.a, c.y, c.s, V, Off, Dth, A.init,
                        As.init, P.a, 0.01, P.y, 1600, 0.7*Scale,
                        10^(-8), FALSE, 10, TRUE)
#
#   Here are the answers you should find (line 7 of Table 1)
#
#   Age.dx = 5, Year.dx = 5, Shock.dx = 15
#   c.a = 19, c.y = 23, c.s = 9
#   Lambda.a = 0.01, Lambda.y = 1600, Lambda.s = 0.7
#   Tr = 283.9, Dev = 9749.7, Bic = 12307.5
#
#   ========================================================
#
#   Final graphical output
#
#   Calculate fitted surface, fitted smooth ie less fitted shocks,
#   and fitted shocks for fitted surface for smooth for Model 4
#
A.smooth2 <- Ans.Scale$A.smooth
A.shock2 <- Ans.Scale$A.shock
Ita2 <- Ans.Scale$Ita - Off
Ita.smooth2 <- B.a %*% matrix(A.smooth2, nrow = ncol(B.a),
                              ncol = ncol(B.y)) %*% t(B.y)
Ita.shock2 <- Ita2 - Ita.smooth2
#
#   Plot output: this is essentially Figure 4 with the addition of the
#   simple shock scaling
#
Plot.year <- function(Yr){
  Yr.row <- Yr - 1899
  Obs <- log(Dth/Exp)[ ,Yr.row]
  plot(Age, Obs - Ita.smooth2[ ,Yr.row], cex = 0.8, 
       ylab = "Mortality shock", main = paste("Mortality shock -", Yr))
  lines(Age, Ita[ ,Yr.row] - Ita.smooth2[ ,Yr.row], col = "red", lwd = 2)
  lines(Age, Ita1[ ,Yr.row] - Ita.smooth2[ ,Yr.row], col = "green", lwd = 2)
  lines(Age, Ita2[ ,Yr.row] - Ita.smooth2[ ,Yr.row], col = "blue", lwd = 2)
  if(Yr == 1918) legend(42, 1.3, c("Alpha = 0", "Alpha = 1", "Alpha = 3.5"),
       lty = 1, lwd = 2, col = c("red", "green", "blue"), cex = 1)
  abline(h = 0)
}
#
par(mfrow = c(2,2))
Plot.year(1918)
Plot.year(1919)
Plot.year(1923)
Plot.year(1944)
#
#     Final perspective plots showing fitted surface, fitted smooth and
#     fitted shocks (Figure 5)
#
par(mfrow = c(1,1))
persp(Age, Year, Ita2, theta = -45, phi = 30, ticktype = "detailed",
      zlab = "log(mortality)", col = "lightblue",
      main = "Smooth + Shocks")
#
persp(Age, Year, Ita.smooth2, theta = -45, phi = 30, ticktype = "detailed",
      zlab = "log(mortality)", col = "lightblue", main = "Smooth")
# 
persp(Age, Year, Ita.shock2, theta = -45, phi = 30, ticktype = "detailed",
      zlab = "log(mortality)", col = "lightblue", main = "Shocks")
