#     
#     File name: Functions.r
#
#     ================================================
#
#      Function to construct a block diagonal matrix from
#      a list of matrices or vectors.  x is a list of matrices.
#      NULLs and scalars are allowed in the list.
#   
dbind <- function(x){
  len.x <- length(x)
  dim.x <- do.call("cbind", lapply(x, function(x)
  if(length(x) == 0) 0 else {
    if(length(x) == 1) c(1, 1) else dim(x)
  }
  ))
  dim.res <- c( sum(dim.x[1,  ]), sum(dim.x[2,  ]) )
  res <- array(0, dim.res)
  res.ends <- cumsum(dim.x[1,  ])
  res.ends2 <- cumsum(dim.x[2,  ])
  res.starts <- c(1, res.ends[ - len.x] + 1)
  res.starts2 <- c(1, res.ends2[ - len.x] + 1)
  for(i in seq(len.x))
  if(dim.x[1, i] > 0) res[res.starts[i]:res.ends[i],
                          res.starts2[i]:res.ends2[i]] <- x[[i]]
  res
}
#
#     ======================================================
#
#     Compute row tensor of two matrices
#
#     Arguments: X. and Y. =  matrices
#
#     Output: the row tensor of X. and Y.
#
Rten <- function(X., Y.){
  one.X <- matrix(1, nrow = 1, ncol = ncol(X.))
  one.Y <- matrix(1, nrow = 1, ncol = ncol(Y.))
  kronecker(X., one.Y) * kronecker(one.X, Y.)
}
#
#     ==============================================================
#
#     Compute B-spline regression matrix.
#
#     Arguments
#
#     X. = domain of vector
#     XL. = left hand knot
#     XR. = right hand knot
#     NDX. = number of intervals in domain
#     BDEG. = degree of B-spline (cubic = 3, etc)
#
#     Output: B-spline regression matrix
#	
bspline <- function(X., XL., XR., NDX., BDEG.){
  dx <- (XR. - XL.)/NDX.
  knots <- seq(XL. - BDEG.*dx, XR. + BDEG.*dx, by=dx)
  B <- spline.des(knots,X.,BDEG.+1,0*X.)$design
  B
}
#
#     For given vector x, a knot is placed at min(x) age.  The function
#     finds the position of the largest knot and the number of internal 
#     gaps between knots for a given gap DX.
#
#     Arguments: X. = x values
#                DX. = interval between knots
#     Output: X.r, maximum knot position
#           : Ndx, the number of internal gaps between knots
#           
#
Setup.X <- function(X., DX.){
   x <- 1:length(X.)
   x.r <- max(seq(1, length(x), by = DX.))
   if(x.r < x[length(x)]) x.r <- x.r + DX.
   ndx <- (x.r - 1)/DX.
   return(list(X.r = x.r, Ndx = ndx))
}
#
#     Computes B and D'D
#
#     Arguments: as bspline( ).  Also PORD. = order of the penalty
#
setup.P <- function(X., XL., XR., BDEG., PORD., NDX.){
  B <- bspline(X., XL., XR., NDX., BDEG.)
  D <- diag(ncol(B))
  for (k in 1:PORD.) D <- diff(D)
  DtD <- t(D) %*% D
  return(list(B = B, DtD = DtD))
}
#
#      GLAM computation of (A2xA1)'W(A2xA1)
#      Code for section (i) in appendix to paper
#
#      Arguments:
#
#      W.  = weight matrix
#      C1. = the number of columns in a marginal regression matrix
#            (and C2.)
#      GB1. = the row tensor of a marginal regression matrix
#            (and GB2.)      
#
Turbo.BWB <- function(W., C1., C2., GB1., GB2.){  
  Ans <- t(GB1.) %*% W. %*% GB2.
  dim(Ans) <- c(C1., C1., C2., C2.)
  Ans <- matrix(aperm(Ans, c(1, 3, 2, 4)), nrow = C1. * C2.)
  Ans
}
#
#      GLAM computation of (I2xA1.cap)'W(A2xA1)
#      Code for section (ii) in appendix to paper
#
#      Arguments:
#
#      W.  = weight matrix, n.a by n.y
#      Ca. = the number of columns in B.a
#      Cy. = the number of columns in B.y
#      Cs. = the number of columns in \breve B.a
#      Ny. = number of rows of B.y
#      by. = vec(B.y)
#      GBaBs. = the row tensor of B.a and B.s
#
Turbo.BWB21 <- function(W., by., GBaBs., Ca., Cy., Cs., Ny.){
  One <- rep(1, Cy.)
  stackW <- kronecker(One, t(W.))
  Ans <- by. * stackW %*% GBaBs.
  dim(Ans) <- c(Ny., Cy., Cs., Ca.)
  Ans <- matrix(aperm(Ans, c(3, 1, 4, 2)), nrow = Cs. * Ny.)
  Ans
}
#
#     GLAM computation of Inverse of (I2xA1)'W(I2xA1) + P
#     Code for section (iii) in appendix to paper
#
#      Arguments:
#
#      W.  = weight matrix, n.a by n.y
#      Cs. = the number of columns in \breve B.a
#      Ny. = number of rows of B.y
#      GB. = the row tensor of A1
#      Lambda. = vector of length Ny. containing values of lambda's 
#
Turbo.BWB22.Inv <- function(W., GB., Cs., Ny., Lambda.){
  GBW <- t(GB.) %*% W.
  Block <- matrix(0, Cs. * Ny.,  Cs. * Ny.)
  Diag <- diag(rep(Lambda., each = Cs.))
  for (j in 1:Ny.){
    r <- (1:Cs.) + (j-1) * Cs.
    Block[r, r] <- chol2inv(chol(matrix(GBW[ , j], Cs., Cs.) +
                                  Diag[r,r]))
  }
  Block
}
#
#   Inverse of a partition matrix [A    :    B  ]
#                                 [t(B) : C.Inv}]
#
Partition.Inverse <- function(A., B., C.Inv.){
  G <- B. %*% C.Inv.
  X <- chol2inv(chol((A. - G %*% t(B.))))
  Y <- -X %*% G
  Z <- C.Inv. - t(G) %*% Y
  row.1 <- cbind(X, Y)
  row.2 <- cbind(t(Y), Z)
  list(Inverse = rbind(row.1, row.2))
}
#
#     Function to calculate SE's
#
#      C1. = the number of columns in a marginal regression matrix
#            (and C2.)
#      B1. = the row tensor of a marginal regression matrix
#            (and B2.)
#      QpP. = B'WB + P (output from the main calculation routine)
#
Turbo.SE <- function(C1., C2., B1., B2., QpP.){
  G <- chol2inv(chol(QpP.))
  G <- array(G, c(C1.,C2.,C1.,C2.))
  G <- aperm(G, c(1,3,2,4))
  G <- array(G, c(C1.^2, C2.^2))
  Dim <- c(nrow(B1.), nrow(B2.)) 
  array(sqrt(B1. %*% G %*% t(B2.)), Dim)
}
#
#     Update function: single GLAM iteration of penalised scoring
#     algorithm for simple 2d-smooth
#
#     Arguments
#
#     A1. = matrix, marginal regression (and A2.)
#     B1. = matrix, row tensor of A1. (and B2.)
#     C1. = scalar, number of columns in A1. (and C2.)
#     V.  = matrix, data present/missing
#     OFF. = offset 
#     D.  = matrix, observed deaths
#     A.  = a matrix, ncol = 1, current estimate of regression
#           coefficients
#     P.  = penalty matrix
#
#     Output
#
#     Estimate of the smooth coefficients
#
Update <- function(A1., A2., B1., B2., C1., C2., V., OFF., D., A., P.){
  a <- matrix(A., nrow = C1., ncol = C2.) 
  Ita <- OFF. + A1. %*% a %*% t(A2.)
  Mu <- exp(Ita)
  Wt <- Mu * V.
  Bt.W.B <- Turbo.BWB(Wt, C1., C2., B1., B2.)
  Rhs1 <- matrix(t(A1.) %*% (V. * (D. - Mu)) %*% A2., ncol = 1)
  Rhs <- Bt.W.B %*% A. + Rhs1
  a <- chol2inv(chol(Bt.W.B + P.)) %*% Rhs
  matrix(a, ncol = 1)
 }
#
#     Update function: single iteration of penalised scoring
#     algorithm for shock model
#
#     Arguments
#
#     A1.  = matrix, marginal regression (and A2.)
#     Asm. = matrix, marginal regression matrix for shocks
#     B1.  = matrix, row tensor of A1. (and B2.)
#     Bs.  = matrix, row tensor of Asm.
#     Bas. = matrix, row tensor of A1. (age) and As. (shocks)
#     C1.  = scalar, number of columns in A1. (and C2.)
#     Cs.  = scalar, number of columns in As. 
#     V.   = matrix, data present/missing
#     OFF. = offset 
#     D.   = matrix, observed deaths
#     A.   = a matrix, ncol = 1, current estimate of regression
#           coefficients
#     Asc. = a matrix, ncol = 1, current estimate of regression
#           coefficients for shocks
#     P.   = penalty matrix for smooth surface
#     L.   = smoothing parameter for shocks
#
#     Output
#
#     Estimates of the smooth and the shock coefficients
#
#
Update.Shock <- function(A1., A2., Asm., B1., B2., Bs., Bas., C1., C2., Cs.,
                         V., OFF., D., A., Asc., P., L.){
  a <- matrix(A., nrow = C1., ncol = C2.)
  ny <- nrow(A2.)
  asc <- matrix(Asc., nrow = Cs., ncol = ny)
  Ba <- A1. %*% a %*% t(A2.) + Asm. %*% asc
  Ita <- OFF. + Ba
  Mu <- exp(Ita)
  Wt <- Mu * V.
  Bt.W.B <- Turbo.BWB(Wt, C1., C2., B1., B2.) + P.
  Bt.W.B.21 <- Turbo.BWB21(Wt, c(A2.), Bas., C1., C2., Cs., ny)
  Bt.W.B.22 <- Turbo.BWB22.Inv(Wt, Bs., Cs., nrow(A2.), L.)
  InverseBWB <- Partition.Inverse(Bt.W.B, t(Bt.W.B.21), Bt.W.B.22)
  RHS <- Wt * (Ba + D./Mu - 1)
  RHS <- c( c(t(A1.) %*% RHS %*% A2.), c(t(Asm.) %*% RHS) )
  RHS <- matrix(RHS, ncol = 1)
  a <- c(InverseBWB$Inverse %*% RHS)
  a.smooth <- matrix(a[1:(C1.*C2.)], ncol = 1)
  a.shock <- matrix(a[(C1.*C2.+1) : length(a)], ncol = 1)
  return(list(A.smooth = a.smooth, A.shock = a.shock))
 }
#
#     As Update.Shock but with additional output
#
Update.Shock.Out <- function(A1., A2., Asm., B1., B2., Bs., Bas.,
                             C1., C2., Cs., V., OFF., D., A., Asc., P., L.){
  a <- matrix(A., nrow = C1., ncol = C2.)
  ny <- nrow(A2.)
  asc <- matrix(Asc., nrow = Cs., ncol = ny)
  Ba <- A1. %*% a %*% t(A2.) + Asm. %*% asc
  Ita <- OFF. + Ba
  Mu <- exp(Ita)
  Wt <- Mu * V.
  Bt.W.B <- Turbo.BWB(Wt, C1., C2., B1., B2.) + P.
  Bt.W.B.21 <- Turbo.BWB21(Wt, c(A2.), Bas., C1., C2., Cs., ny)
  Bt.W.B.22 <- Turbo.BWB22.Inv(Wt, Bs., Cs., nrow(A2.), L.)
  InverseBWB <- Partition.Inverse(Bt.W.B, t(Bt.W.B.21), Bt.W.B.22)
  RHS <- Wt * (Ba + D./Mu - 1)
  RHS <- c( c(t(A1.) %*% RHS %*% A2.), c(t(Asm.) %*% RHS) )
  RHS <- matrix(RHS, ncol = 1)
  a <- c(InverseBWB$Inverse %*% RHS)
  a.smooth <- matrix(a[1:(C1.*C2.)], ncol = 1)
  a.shock <- matrix(a[(C1.*C2.+1) : length(a)], ncol = 1)
  return(list(A.smooth = a.smooth, A.shock = a.shock, Ita = Ita,
              Mu = Mu, InverseBWB = InverseBWB$Inverse))
 }
#
#      Estimate regression coefficients.  This calls the update function
#      until convergence.  The user can set convergence criterion and
#      maximum number of iterations.  Monitoring of output can be
#      switched off/on.
#
#      Arguments
#
#      As Update().  Also
#      A.INIT = initial estimate of a
#      TOL. = tolerance, set to 10^(-6) or otherwise
#      MONITOR. = monitoring, FALSE = off, TRUE = on
#      MAX.ITER. = maximum no of iterations, set to 20 or otherwise
#
#      Output
#
#      Estimate of coefficients for given lambda for smooth model
#
Estimate <- function(A1., A2., B1., B2., C1., C2., V., OFF., D.,
                     A.INIT., P., TOL., MONITOR., MAX.ITER.){
  a <- A.INIT.
  a.old <- 10
  iter <- 0
  tol = 1
  if (MONITOR. == TRUE) cat("Iter     a[1]            tol", "\n")
  while(tol > TOL. && iter < MAX.ITER.) {
    iter <- iter + 1
    a <- Update(A1., A2., B1., B2., C1., C2., V., OFF., D., a, P.)
    tol <- max(abs(a - a.old)/abs(a))
    a.old <- a
    if (MONITOR. == TRUE) cat(iter, a[11], tol, "\n")
  }
  if (iter == MAX.ITER.) warning("Failed to converge")
  matrix(a, ncol = 1)
}
#
#      Estimate regression coefficients.  This calls the update function
#      until convergence.  The user can set convergence criterion and
#      maximum number of iterations.  Monitoring of output can be
#      switched off/on.
#
#      Arguments
#
#      As Update.Shock().  Also
#      A.INIT. = initial estimate of a.smooth
#      A.INIT. = initial estimate of a.smooth
#      TOL. = tolerance, set to 10^(-6) or otherwise
#      MONITOR. = monitoring, FALSE = off, TRUE = on
#      MAX.ITER. = maximum no of iterations, set to 20 or otherwise
#
#      Output
#
#      Estimate of coefficients for given lambda and lambda.s
#      for shock model
#
Estimate.Shock <- function(A1., A2., Asm., B1., B2., Bs., Bas.,
                           C1., C2., Cs., V., OFF., D., A.INIT., As.INIT.,
                           P., L., TOL., MONITOR., MAX.ITER.){
  a.smooth <- A.INIT.
  a.shock <- As.INIT.
  a.smooth.old <- 10
  a.shock.old <- 10
  iter <- 0
  tol = 1
  if (MONITOR. == TRUE) cat("Iter a.smooth[1] a.shock[1] tol", "\n")
  
  while(tol > TOL. && iter < MAX.ITER.) {
    iter <- iter + 1
    a <- Update.Shock(A1., A2., Asm., B1., B2., Bs., Bas., C1., C2., Cs.,
                      V., OFF., D., a.smooth, a.shock, P., L.)
    a.smooth <- a$A.smooth
    a.shock <- a$A.shock
    tol <- max(abs(a.smooth - a.smooth.old)/abs(a.smooth),
               abs(a.shock - a.shock.old)/abs(a.shock))
    a.smooth.old <- a.smooth
    a.shock.old <- a.shock
    if (MONITOR. == TRUE) cat(iter, a.smooth[11], a.shock[11], tol, "\n")
  }
  if (iter == MAX.ITER.) warning("Failed to converge")
  return(list(A.smooth = matrix(a.smooth, ncol = 1),
              A.shock = matrix(a.shock, ncol = 1)))
}
#
#       Main function: calculates for simple 2d-smoothing
#
#          Calc$a   = estimates of regression coefficients
#          Calc$Hazard = the hazard function = Ba
#          Calc$Tr  = trace of hat matrix
#          Calc$Dev = deviance
#          Calc$Bic = BIC (Bayesian Information Criterion)
#          Calc$BtWBplusP = B'WB + P (needed for SE's)
#
#       Arguments
#
#       A.INIT., X., V., OFF., D., TOL., MONITOR., MAX.ITER. as above.
#       Also
#       P1. = penalty matrix for age
#       LAMBDA1. = smoothing parameter for age
#       P2. = penalty matrix for year
#       LAMBDA2. = smoothing parameter for year
#       TRACE. = used in conjunction with minimization
#       If ms() is used set to TRUE, otherwise FALSE
#
Calc <- function(A1., A2., B1., B2., C1., C2., V., OFF., D., A.INIT., 
                 P1., LAMBDA1., P2., LAMBDA2.,
                 TOL., MONITOR., MAX.ITER., TRACE.){
  lambda1 <- abs(LAMBDA1.)
  lambda2 <- abs(LAMBDA2.)
  P <- lambda1 * P1. + lambda2 * P2.
  a <- Estimate(A1., A2., B1., B2., C1., C2., V., OFF., D., A.INIT., P,
                TOL., MONITOR., MAX.ITER.)
  a <- array(a, c(C1., C2.)) 
  Ita <- OFF. + A1. %*% a %*% t(A2.)
  Mu <- exp(Ita)
  Wt <- Mu * V.
  Bt.W.B <- Turbo.BWB(Wt, C1., C2., B1., B2.)
  BtWBplusP <- Bt.W.B + P
  Rhs1 <- matrix(t(A1.) %*% (V. * (D. - Mu)) %*% A2., ncol = 1)
  Rhs <- Bt.W.B %*% matrix(a, ncol = 1) + Rhs1
  a <- solve(BtWBplusP, Rhs)
  a <- array(a, c(C1., C2.))
  Tr <- sum(diag(solve(BtWBplusP, Bt.W.B)))
  y.init <- D.
  y.init[D.==0] <- 10^(-4)
  Dev <- 2*sum( V. * D.*log(y.init/Mu) )
  Aic <- Dev + 2 * Tr
  Bic <- Dev + log(sum(V.)) * Tr
  if (TRACE. == TRUE) cat(lambda1, lambda2, Bic, "\n")
  return(list(a = a, Ita = Ita, Tr = Tr, Dev = Dev, Aic = Aic,
               Bic = Bic, BtWBplusP = BtWBplusP))
}
#
#       Main function for shock: calculates for smooth + shocks
#
#          Calc$a   = estimates of regression coefficients
#          Calc$Hazard = the hazard function = Ba
#          Calc$Tr  = trace of hat matrix
#          Calc$Dev = deviance
#          Calc$Bic = BIC (Bayesian Information Criterion)
#          Calc$BtWBplusP = B'WB + P (needed for SE's)
#
#       Arguments
#
#       A.INIT., X., V., OFF., D., TOL., MONITOR., MAX.ITER. as above.
#       Also
#       P1. = penalty matrix for age
#       LAMBDA1. = smoothing parameter for age
#       P2. = penalty matrix for year
#       LAMBDA2. = smoothing parameter for year
#       TRACE. = used in conjunction with minimization
#
Calc.Shock <- function(A1., A2., Asm., B1., B2., Bs., Bas., C1., C2., Cs.,
                       V., OFF., D., A.INIT., As.INIT.,
                       P1., LAMBDA1., P2., LAMBDA2., LAMBDA3.,
                       TOL., MONITOR., MAX.ITER., TRACE.){
  lambda1 <- abs(LAMBDA1.)
  lambda2 <- abs(LAMBDA2.)
  lambda3 <- abs(LAMBDA3.)
  P <- lambda1 * P1. + lambda2 * P2.
  a <- Estimate.Shock(A1., A2., Asm., B1., B2., Bs., Bas., C1., C2., Cs.,
                      V., OFF., D., A.INIT., As.INIT., P, LAMBDA3., 
                      TOL., MONITOR., MAX.ITER.)
  Output <- Update.Shock.Out(A1., A2., Asm., B1., B2., Bs., Bas.,
                             C1., C2., Cs., V., OFF., D.,
                             a$A.smooth, a$A.shock, P, lambda3)
  A.smooth <- Output$A.smooth
  A.shock <- Output$A.shock
  Ita <- Output$Ita
  Mu <- Output$Mu
  InverseBWB <- Output$InverseBWB
  BigP <- dbind(list(P, diag(rep(lambda3, each = Cs.))))
  Tr <- C1. * C2. + Cs. * ncol(D.) - sum(diag(InverseBWB %*% BigP))
  y.init <- D.
  y.init[D.==0] <- 10^(-4)
  Dev <- 2*sum( V. * D.*log(y.init/Mu) )
  Aic <- Dev + 2 * Tr
  Bic <- Dev + log(sum(V.)) * Tr
  if (TRACE. == TRUE) cat(lambda1, lambda2, lambda3[1], Bic, "\n")
  return(list(A.smooth = A.smooth, A.shock = A.shock, Ita = Ita, Tr = Tr,
              Dev = Dev, Aic = Aic, Bic = Bic, InverseBWB = InverseBWB))
}

